#include "segment.hpp"
#include "coil.hpp"
#include "glm/ext/matrix_float4x4.hpp"
#include "glm/ext/matrix_transform.hpp"
#include "glm/trigonometric.hpp"
#include <cstring>
#include <iostream>
#include <glm/glm.hpp>
#include <cstdint>

// x, y, z, r, g, b
#define FIELDS_IN_POINT 6
#define DAYS_IN_YEAR (isLeap? 366.0f : 365.0f)

using glm::vec4;
using glm::vec3;
using glm::mat4;
using glm::radians;
using glm::rotate;
using glm::sin;
using glm::cos;
using glm::translate;

using std::endl;
using std::cout;
using std::cerr;

CoilSegment::CoilSegment(
        Coil* coil, 
        uint16_t year, 
        uint16_t height, 
        int shift, 
        uint32_t sliceDetalization, 
        bool isLeap
    ) :
    coil(coil), year(year), 
    height(height), shift(shift), 
    sliceDetalization(sliceDetalization), isLeap(isLeap) {
        init();
        calculate();
    }

void CoilSegment::init () {
    verticesLength = DAYS_IN_YEAR * FIELDS_IN_POINT * (sliceDetalization+2);
    vertices = (double *) calloc(verticesLength, sizeof(double));
    if (vertices == NULL) {
        cerr << "Allocation failed" << endl;
        exit(-1);
    }
    // cout << "Initialized CoilSegment" 
    //     << "(year: " << year 
    //     << ", height: " << height 
    //     << ", shift: " << shift
    //     << ", isLeap: " << isLeap <<")" << endl;
}

void CoilSegment::printVertices() {
    for (uint16_t day = 0; day < DAYS_IN_YEAR; day ++) {
        cout << "Day " << day << ", " <<  
            "x: " << vertices[FIELDS_IN_POINT * day + 0] << ", " <<
            "y: " << vertices[FIELDS_IN_POINT * day + 1] << ", " <<
            "z: " << vertices[FIELDS_IN_POINT * day + 2] << ", " <<
            "r: " << vertices[FIELDS_IN_POINT * day + 3] << ", " <<
            "g: " << vertices[FIELDS_IN_POINT * day + 4] << ", " <<
            "b: " << vertices[FIELDS_IN_POINT * day + 5] << endl;
    }
}

double round_to_presicion( double value, int precision ) {
    const int adjustment = pow(10,precision);
    return floor( value*(adjustment) + 0.5 )/adjustment;
}

double *CoilSegment::calculateSlice() {
    uint32_t size = (/*start and end of a slize*/2 + sliceDetalization) * FIELDS_IN_POINT * sizeof(double);
    double *slice = (double *)malloc(size);
    // Start of a slice
    slice[0] = cos(0);
    slice[1] = sin(0);
    slice[2] = 0;
    slice[3] = 1;
    slice[4] = 1;
    slice[5] = 1;

    double degreeByPoint = -180. / (sliceDetalization + 1);

    for (uint32_t i = 1; i <= sliceDetalization; i ++) {
        uint32_t offset = FIELDS_IN_POINT * i;
        
        slice[offset + 0] = round_to_presicion(cos(radians(degreeByPoint * i)), 5);
        slice[offset + 1] = round_to_presicion(sin(radians(degreeByPoint * i)), 5);
        slice[offset + 2] = 0;
        slice[offset + 3] = 1;
        slice[offset + 4] = 1;
        slice[offset + 5] = 1;
    }

    // End of a slice
    uint32_t endIndex = (size / sizeof(double) - FIELDS_IN_POINT); 
    slice[endIndex + 0] = round_to_presicion(cos(radians(180.)), 5);
    slice[endIndex + 1] = round_to_presicion(sin(radians(180.)), 5);
    slice[endIndex + 2] = 0;
    slice[endIndex + 3] = 1;
    slice[endIndex + 4] = 1;
    slice[endIndex + 5] = 1;

    return slice;
}

void CoilSegment::constructSegment(double *slice) {

    double degreesPerDay = 360.f / DAYS_IN_YEAR;

    for (uint32_t day = 0; day < DAYS_IN_YEAR; day ++) {
        float daysDegree = day * degreesPerDay;
        // TODO: replace 10 to coil->getWidth();
        vec3 daysPosition = vec3(
            cos(radians(daysDegree)) * 200,
            10* daysDegree/360.,
            sin(radians(daysDegree)) * 200
        );

        mat4 transform = mat4(1.);
        // For some reason I cannot combine transform and rotation to one singe matrix.
        // TODO: investigate and fix        
        // transform = translate(
        //     transform,
        //     daysPosition
        // );

        transform = rotate(
            transform,
            radians(daysDegree),
            vec3(
                0.0,
                1.0,
                0.0
            )
        );

        for (uint32_t slicePoint = 0; slicePoint < (2 + sliceDetalization); slicePoint ++) {
            uint32_t slicePointOffset = FIELDS_IN_POINT * slicePoint;
            vec4 point(
                slice[slicePointOffset + 0],
                slice[slicePointOffset + 1],
                slice[slicePointOffset + 2],
                1
            );
            
            point = point*transform  ;
            uint64_t currentPointOffset = day * (2+sliceDetalization) * FIELDS_IN_POINT + slicePointOffset;
            vertices[currentPointOffset + 0] = point.x + daysPosition.x;
            vertices[currentPointOffset + 1] = point.y + daysPosition.y;
            vertices[currentPointOffset + 2] = point.z + daysPosition.z;

            memcpy(
                vertices + currentPointOffset + 3,
                slice + slicePointOffset + 3,
                3
            );
        }
    }
}
// Sample layout of 2 slices with 3 points.
//  |x |y |z |r |g |b |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |
//  0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38
//  |     point       |    point        |     point       |    point        |      point      |    point        |  
//  |                  slice                              |             slice                                   |
//  
void CoilSegment::calculate() {
    double *slice = calculateSlice();

    constructSegment(slice);
    exportSegmentToCsv();

}

double *CoilSegment::getVertices() {
    return vertices;
}

uint16_t CoilSegment::getVerticesLength() {
    return verticesLength;
}

void CoilSegment::printSlice(double *slice) {
    for (uint32_t i = 0; i < sliceDetalization+2; i ++) {
        uint32_t offset = FIELDS_IN_POINT * i;
        cout << "Point ("
                  << slice[offset + 0] << ", "
                  << slice[offset + 1] << ", "
                  << slice[offset + 2] << ", "
                  << ") rgb("
                  << slice[offset + 3] << ", "
                  << slice[offset + 4] << ", "
                  << slice[offset + 5] << ")"
                  << endl;
    }
}

void CoilSegment::exportSliceToCSV(double *slice) {
    cout << "x,y,z" << endl;
    
    for (uint32_t i = 0; i < sliceDetalization+2; i ++) {
        uint32_t offset = FIELDS_IN_POINT * i;
        cout
                  << slice[offset + 0] << ","
                  << slice[offset + 1] << ",0" << endl;
    }
}

void CoilSegment::exportSegmentToCsv() {
    cout << "x,y,z" << endl;

    for (uint16_t day = 0; day < DAYS_IN_YEAR; day ++) {
        uint64_t dayOffset = day * (2+sliceDetalization) * FIELDS_IN_POINT;

        for (uint32_t point_in_slice = 0; point_in_slice < (2+sliceDetalization); point_in_slice ++) {
            uint32_t sliceOffset = point_in_slice * FIELDS_IN_POINT;
            cout << vertices[dayOffset +  sliceOffset + 0] << "," 
                      << vertices[dayOffset +  sliceOffset + 1] << ","
                      << vertices[dayOffset +  sliceOffset + 2] << endl; 
        }
    }
}